import os
import numpy as np
import subprocess
import tempfile
import pandas as pd
from molbar.helper.vector import Vector
import matplotlib.pyplot as plt
from molbar.helper.printer import Printer
from scipy.stats import rankdata

class DoubleBondAnalysis:

    def _get_double_bonds_in_fragment(self, atoms_in_fragment, double_bonds):

        return [db for db in double_bonds if db[0] in atoms_in_fragment and db[1] in atoms_in_fragment]

    def _check_fragments_for_double_bonds(self, original_fragment, duplicate_fragment, atoms_in_fragment_original, atoms_in_fragment_duplicate):

        original_double_bonds = self.data_original["double_bonds"]

        duplicate_double_bonds = self.data_duplicate["double_bonds"]

        original_double_bonds = self._get_double_bonds_in_fragment(atoms_in_fragment_original, original_double_bonds)

        duplicate_double_bonds = self._get_double_bonds_in_fragment(atoms_in_fragment_duplicate, duplicate_double_bonds)

        if len(original_double_bonds) == 0 and len(duplicate_double_bonds) == 0:

            return None, []

        db_indices_original, db_types_original = self._get_double_bond_indices(self.data_original, original_double_bonds)

        db_indices_duplicate, db_types_duplicate = self._get_double_bond_indices(self.data_duplicate, duplicate_double_bonds)

        if len(db_indices_original) == 0 and len(db_indices_duplicate) == 0:

            return None, []
        
        if len(db_indices_original) != len(db_indices_duplicate):
            
            return None, []
        
        comparisons = [self._compare_double_bond_configuration(original_fragment, duplicate_fragment, db_index_original, db_index_duplicate) for db_index_original, db_index_duplicate in zip(db_indices_original, db_indices_duplicate)]

        cause = None

        changed_db_atoms = []

        if any(comparisons):

            cause = "change_double_bond_configuration"

            bo_matrix_original = self._get_bond_orders(type="original")

            for i, comparison in enumerate(comparisons):

                if comparison:

                    changed_db_atoms.append(db_indices_original[i][:2])
                    
                    self.changed_db_types.append(db_types_original[i])

                    self.changed_db_cid.append(self.data_original["CID"])

                    wbo = bo_matrix_original[db_indices_original[i][0], db_indices_original[i][1]] 

                    self.changed_db_wbos.append(wbo)

        return cause, changed_db_atoms



    def _get_double_bond_indices(self, data_dict: dict, db_bonds: list):

        """
        Get the atomic indices of the double bonds (bond indices and adjacent atom indices) and the double bond types.

        Returns:
            all_indices (list): List of lists of the atomic indices of the double bonds (bond indices and adjacent atom indices)
        """

        elements = data_dict["elements"]

        charges = data_dict["atomic_numbers"]

        priorities = data_dict["priorities"]

        all_indices = []

        db_types = []

        for bond in db_bonds:

            ith_vsepr_class = data_dict["vsepr_classes"][bond[0]]

            jth_vsepr_class = data_dict["vsepr_classes"][bond[1]]

            if ith_vsepr_class not in ["bent", "trigonal_planar"]:

                continue

            if jth_vsepr_class not in ["bent", "trigonal_planar"]:

                continue

            ith_adjacent_atoms = [x for x in data_dict["graph"][bond[0]] if x != bond[1]]

            jth_adjacent_atoms = [x for x in data_dict["graph"][bond[1]] if x != bond[0]]

            indices = bond + ith_adjacent_atoms + jth_adjacent_atoms

            db_type = self._get_double_bond_types(bond, ith_adjacent_atoms, jth_adjacent_atoms, elements, charges)

            db_types.append(db_type)

            all_indices.append(indices)

        if len(all_indices) == 0:

            return [], []

        db_priorities = [list(priorities[indices]) for indices in all_indices]

        padded_db_priorities = self.get_padded_lists(db_priorities)

        sorting_indices = self.argsort_lists(padded_db_priorities)

        sorted_all_indices = [all_indices[sorting_index] for sorting_index in sorting_indices]

        sorted_db_types = [db_types[sorting_index] for sorting_index in sorting_indices]

        return sorted_all_indices, sorted_db_types
    
    def _get_double_bond_types(self, double_bond: list, ith_adjacent_atoms: list, jth_adjacent_atoms: list, elements: list, charges: list):
        """
        Get the double bond type.

        Args:
            double_bond (list): double bond indices
            ith_adjacent_atoms (list): adjacent atom indices of first atom of double bond
            jth_adjacent_atoms (list): adjacent atom indices of second atom of double bond
            elements (list): list of elements 
            charges (list): list of charges

        Returns:
            db_type: double bond type
        """

        db_elements = [elements[double_bond[0]], elements[double_bond[1]]]

        ith_elements = sorted([elements[x] if elements[x] == "O" else "R" for x in ith_adjacent_atoms])

        jth_elements =  sorted([elements[x] if elements[x] == "O" else "R" for x in jth_adjacent_atoms])

        ith_charges = sum([charges[double_bond[0]]] + [charges[x] for x in ith_adjacent_atoms])

        jth_charges = sum([charges[double_bond[1]]] + [charges[x] for x in jth_adjacent_atoms])

        if ith_charges > jth_charges:

            ith_elements, jth_elements = jth_elements, ith_elements

            db_elements[0], db_elements[1] = db_elements[1], db_elements[0]

        if len(ith_elements) == 1 and len(jth_elements) == 1:

            db_type = ith_elements[0] + db_elements[0]  + "=" + db_elements[1] + jth_elements[0]

        elif len(ith_elements) == 1 and len(jth_elements) == 2:

            db_type = ith_elements[0] + db_elements[0] + "=" + db_elements[1] + "("+jth_elements[0] +")" + jth_elements[1]

        elif len(ith_elements) == 2 and len(jth_elements) == 1:

            db_type = ith_elements[1] + "("+ith_elements[0] +")" + db_elements[0] + "=" + db_elements[1] +  jth_elements[0]

        elif len(ith_elements) == 2 and len(jth_elements) == 2:

            db_type = ith_elements[1] + "("+ith_elements[0] +")" + db_elements[0] + "=" + db_elements[1] + "("+jth_elements[0] +")" + jth_elements[1]

        else:

            db_type = "unknown"

        return db_type


    
    def _get_double_configuration(self, db_index, data_dict, coordinates, atoms_in_fragment):

        ith_adjacent_atoms = [x for x in data_dict["graph"][db_index[0]] if x != db_index[1]]

        jth_adjacent_atoms = [x for x in data_dict["graph"][db_index[1]] if x != db_index[0]]

        ith_priorities = data_dict["priorities"][ith_adjacent_atoms]

        if len(ith_priorities) != len(set(ith_priorities)):

            return 0

        jth_priorities = data_dict["priorities"][jth_adjacent_atoms]

        if len(jth_priorities) != len(set(jth_priorities)):

            return 0
        
        selected_ith_adjacent_atom = ith_adjacent_atoms[np.argmax(ith_priorities)]

        selected_jth_adjacent_atom = jth_adjacent_atoms[np.argmax(jth_priorities)]

        global_selected_indices = [selected_ith_adjacent_atom, db_index[0], db_index[1], selected_jth_adjacent_atom] 

        try:

            selected_indices = [np.where(atoms_in_fragment == x)[0][0] for x in global_selected_indices]

        except IndexError:

            return 0

        if len(selected_indices) != 4:

            return 0

        selected_coordinates = coordinates[selected_indices]

        dihedral = np.abs(Vector(selected_coordinates).calculate_dihedral())

        if dihedral < 45:

            return 1
        
        elif dihedral > 135:

            return -1

        else:

            return 0
    
    def _compare_double_bond_configuration(self, fragment_original, fragment_duplicate, db_index_original, db_index_duplicate):

        configuration_original = self._get_double_configuration(db_index_original, self.data_original, np.array(fragment_original["final_geometry"]), np.array(fragment_original["atoms_in_fragment"])-1)

        configuration_duplicate = self._get_double_configuration(db_index_duplicate, self.data_duplicate, np.array(fragment_duplicate["final_geometry"]), np.array(fragment_duplicate["atoms_in_fragment"])-1)

        if configuration_original != configuration_duplicate:

            return True
        
        else:

            return False


    def _get_sorted_db_spectra(self, all_indices, db_types, cm_matrix, priorities):

        max_length = 0

        spectra = []

        spectra_str = []

        for global_db_indices in all_indices:
 
            local_cm_matrix = cm_matrix[np.ix_(global_db_indices, global_db_indices)]

            diagonal = rankdata(priorities[global_db_indices], method="min")

            np.fill_diagonal(local_cm_matrix, diagonal)

            spectrum = np.round(np.linalg.eigvalsh(local_cm_matrix)*100).astype(int)

            if len(spectrum) > max_length:

                max_length = len(spectrum)  

            spectra_str.append(" ".join([str(x) for x in spectrum]))

            spectra.append(list(spectrum))

        padded_spectra = [spectrum + [0]*(max_length-len(spectrum)) for spectrum in spectra]

        sorting_indices = self.argsort_lists(padded_spectra)

        sorted_spectra = [spectra_str[sorting_index] for sorting_index in sorting_indices]

        sorted_db_indices = [all_indices[sorting_index] for sorting_index in sorting_indices]

        sorted_db_types = [db_types[sorting_index] for sorting_index in sorting_indices]

        return sorted_db_indices, sorted_spectra, sorted_db_types

    def _read_wbo_file(self, n_atoms, file_path):

        bo_matrix = np.zeros((n_atoms, n_atoms))

        with open(file_path, 'r') as file:
            for line in file:
                # Split the line into three values
                atomic_index1, atomic_index2, bond_order = map(float, line.split())
                
                # Append the values to the data list as a tuple
                bo_matrix[int(atomic_index1)-1, int(atomic_index2)-1] = bond_order

                bo_matrix[int(atomic_index2)-1, int(atomic_index1)-1] = bond_order

        return bo_matrix

    def _get_bond_orders(self, type="original"):
        if type == "original":

            data_dict = self.data_original

        else:

            data_dict = self.data_duplicate

        coordinates = data_dict["coordinates"]

        elements = data_dict["elements"]

        # Create a temporary directory
        random_dir = tempfile.mkdtemp()

        n_atoms = len(elements)

        dir_path = os.path.dirname(os.path.realpath(__file__))

        try:
            # Change into the random directory
            os.chdir(random_dir)

            Printer(n_atoms, "", coordinates, elements, "coord.xyz").print()

            # Run a command (replace "your_command" with the actual command you want to run)
            subprocess.run(["xtb coord.xyz > xtb.out 2> xtb.err"], shell=True)

            bo_matrix = self._read_wbo_file(n_atoms, "wbo")

        finally:
            # Change back to the original directory
            os.chdir(dir_path)

            # Remove the random directory and its contents
            subprocess.run(["rm", "-r", random_dir])

        return bo_matrix
    
    def get_db_plot(self):

        df_bond = pd.DataFrame({"CID": self.changed_db_cid, "db_type":self.changed_db_types, "wbo":self.changed_db_wbos})

        df_bond.to_pickle(os.path.join(self.results_path, "double_bond_analysis.pkl"))

        plt.style.use(self.style_path)

        db_types = df_bond["db_type"].unique()

        dict_db_types = {db_type: [] for db_type in db_types}

        for db_type, wbo in zip(df_bond["db_type"], df_bond["wbo"]):

            dict_db_types[db_type].append(wbo)

        data = {}

        other = []

        for key, wbos in dict_db_types.items():

            if key in ["R(R)N=C(O)R", "R(R)C=C(R)R"]:

                data[key] = wbos

            else:

                other += wbos

        data["other"] = other

        values = list(data.values())

        plt.figure(figsize=(12,9))

        plt.hist(values, 100, histtype='bar', stacked=True, label=list(data.keys()))
        plt.legend(prop={'size': 12}, bbox_to_anchor=(0.9, 1), loc='upper left')
        plt.xlabel('Wiberg Bond Order', fontsize=20)
        plt.ylabel('Frequency', fontsize=20)
        plt.xticks(fontsize=20)
        plt.yticks(fontsize=20)
        plt.grid(False)
        #plt.axvline(x=1.1, color='r', linestyle='-')
        plt.savefig(os.path.join(self.figure_path, "double_bond_analysis.png"), dpi=300, bbox_inches='tight')