import pandas as pd
import time
import sys
import itertools
import os
import numpy as np
import matplotlib.pyplot as plt
import re
import threading
import shutil
from modules._inchi_exclusive import InChIExclusive
from modules._molbar_exclusive import MolBarExclusive

class FindDuplicates(InChIExclusive, MolBarExclusive):

    def __init__(self, path_to_df, figures_path, results_path, style_path):
        
        print("#"*97)
        print("#"*40 + " Duplication Test " + "#"*39)
        print("#"*97)

        self.doneloading = False

        t = threading.Thread(target=self._animation)

        t.start()

        self.figure_path = figures_path

        self.results_path = results_path

        if os.path.exists(self.results_path):

            shutil.rmtree(self.results_path)

        os.mkdir(self.results_path)

        self.report_path = os.path.join(self.results_path, "report.txt")

        self.style_path = style_path

        df_total = pd.read_pickle(path_to_df)

        self.doneloading = True

        print("\n")

        pattern = re.compile(r'\b\d+\b')

        df_total["CID"] = df_total["comment"].apply(lambda x: int(pattern.findall(x)[0]))

        self.df_total = df_total.copy()

    def find_duplicates(self):

        df_total = self.df_total.copy()
        
        df_total["duplicate_molbar"] = df_total["MolBar"].duplicated()

        df_total["duplicate_inchi"] = df_total["InChI"].duplicated()

        df_total["duplicate_smiles"] = df_total["SMILES"].duplicated()

        columns = ["header", "version", "molecular_formula", "topology", "ha_topology", "topography", "chirality"]

        df_total.loc[:, columns] = df_total["MolBar"].str.split("|", expand=True).values

        df_differences = df_total.loc[df_total["duplicate_molbar"] != df_total["duplicate_inchi"]]

        self.df_differences = df_differences.copy()

        self.df_total = df_total.copy()

        self._get_duplicates_plot()

        self._duplicate_due_to_molbar()

        self._duplicate_due_to_inchi()

        self._do_statistics_inchi_exclusive_duplicates()

        self._create_report()

        self._analyze_inchi_exclusive_duplicates()

        self.print_report()


    def _get_duplicates_plot(self):

        """
        Generate a plot of the number of duplicates in the dataset.

        Parameters:
        -----------
        df: pandas.DataFrame

        Returns:
        --------
        None
        """

        plt.style.use(self.style_path)

        data = [self.df_total["duplicate_molbar"].sum(), self.df_total["duplicate_smiles"].sum(), self.df_total["duplicate_inchi"].sum()]

        fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(12, 9))

        fig.subplots_adjust(hspace=0.05)  # adjust space between axes

        p = 1e6

        ax1.ticklabel_format(axis='y', style='sci', scilimits=(6, 6))
        ax1.set_ylim(3.6*p, 4*p)  # outliers only

        ax1.axhline(y=len(self.df_total["duplicate_molbar"]), color='black', linestyle='--')

        #ax2.ticklabel_format(axis='y', style='sci', scilimits=(4, 4))
        # plot the same data on both axes
        ax2.set_ylim(0, 0.25)  # most of the data
        ax2.bar(0, data[0]/p, width=0.5)
        ax2.bar(1, data[1]/p, width=0.5)
        ax2.bar(2, data[2]/p, width=0.5)
        ax2.set_xticks([0, 1, 2], ["MolBar", "SMILES", "InChI"], fontsize=20)
        ax2.set_yticks([0.1, 0.2], ["0.1", "0.2"])

        formatted_data0 = '{:,}'.format(data[0])
        formatted_data2 = '{:,}'.format(data[2])
        formatted_data1 = '{:,}'.format(data[1])
        formatted_data_all = '{:,}'.format(len(self.df_total))

        ax1.annotate(formatted_data_all, xy=(1, len(self.df_total)), ha='center', va='bottom', fontsize=20)
        ax2.annotate(formatted_data0, xy=(0, data[0]/p), ha='center', va='bottom', fontsize=20)
        ax2.annotate(formatted_data1, xy=(1, data[1]/p), ha='center', va='bottom', fontsize=20)
        ax2.annotate(formatted_data2, xy=(2, data[2]/p), ha='center', va='bottom', fontsize=20)
        fig.text(0.02, 0.5, 'Number of Duplicates', va='center', rotation='vertical', fontsize=20)
        # zoom-in / limit the view to different portions of the data

        # hide the spines between ax and ax2
        ax1.spines.bottom.set_visible(False)
        ax2.spines.top.set_visible(False)
        ax1.xaxis.tick_top()
        ax1.tick_params(labeltop=False)  # don't put tick labels at the top
        ax2.xaxis.tick_bottom()

        d = .5  # proportion of vertical to horizontal extent of the slanted line
        kwargs = dict(marker=[(-1, -d), (1, d)], markersize=12,
                    linestyle="none", color='k', mec='k', mew=1, clip_on=False)
        ax1.plot([0, 1], [0, 0], transform=ax1.transAxes, **kwargs)
        ax2.plot([0, 1], [1, 1], transform=ax2.transAxes, **kwargs)

        plt.savefig(os.path.join(self.figure_path, "duplicate_test.png"), dpi=300, bbox_inches='tight')

    def _duplicate_due_to_molbar(self):

        df_original = self.df_total[self.df_total["duplicate_inchi"] == False]

        df_molbar_exclusive = self.df_differences.loc[self.df_differences["duplicate_molbar"] == True]

        df_molbar_exclusive = df_molbar_exclusive.merge(df_original[["CID", "molfile", "MolBar", "InChI", "coordinates", "elements"]], left_on="MolBar", right_on="MolBar", how="left", suffixes=("_duplicate", "_original"))

        df_molbar_exclusive = df_molbar_exclusive.loc[df_molbar_exclusive["CID_original"] != df_molbar_exclusive["CID_duplicate"]]

        df_molbar_exclusive.to_pickle(os.path.join(self.results_path, "molbar_exclusive_duplicates.pkl"))

        self.df_molbar_exclusive = df_molbar_exclusive.copy()

        self.generate_pdf(df_molbar_exclusive, os.path.join(self.results_path, "molbar_exclusive_duplicates.pdf"))

        self._print_molecules_to_dir(df_molbar_exclusive, os.path.join(self.results_path, "molbar_exclusive_duplicates"), "MolBar exlusive duplicates")

    def _duplicate_due_to_inchi(self):

        df_original = self.df_total[self.df_total["duplicate_inchi"] == False]

        df_inchi_exclusive = self.df_differences.loc[self.df_differences["duplicate_inchi"] == True]

        df_inchi_exclusive = df_inchi_exclusive.merge(df_original[["CID", "molfile", "MolBar", "InChI", "coordinates", "elements", "molecular_formula", "topology", "ha_topology", "topography", "chirality"]], left_on="InChI", right_on="InChI", how="left", suffixes=("_duplicate", "_original"))
        
        df_inchi_exclusive = df_inchi_exclusive.loc[df_inchi_exclusive["CID_original"] != df_inchi_exclusive["CID_duplicate"]]

        columns = ['topology', 'ha_topology', 'topography', 'chirality']

        for column in columns:
            # Analye if original and duplicate of one column are different
            df_inchi_exclusive[f"{column}_change"] = df_inchi_exclusive[f"{column}_original"] != df_inchi_exclusive[f"{column}_duplicate"]

        df_inchi_exclusive["change_molecular_formula"] = df_inchi_exclusive["molecular_formula_original"] != df_inchi_exclusive["molecular_formula_duplicate"]

        df_inchi_exclusive["constitutional_isomerism"] = False

        # If heavy atom topology is different, but same molecular formula then it is a constitutional isomer
        df_inchi_exclusive.loc[(df_inchi_exclusive["ha_topology_change"]) & (~df_inchi_exclusive["change_molecular_formula"]), "constitutional_isomerism"] = True

        df_inchi_exclusive["tautomerism"] = False
        # If topology is different, but same molecular formula and same heavy atom topology then it is a tautomer
        df_inchi_exclusive.loc[(df_inchi_exclusive["topology_change"]) & (~df_inchi_exclusive["constitutional_isomerism"])  & (~df_inchi_exclusive["change_molecular_formula"]), "tautomerism"] = True

        df_inchi_exclusive["change_relative_configuration"] = False
        # If topography barcode is different, but same molecular formula, same heavy atom topology and same topology then it is a configurational isomer
        df_inchi_exclusive.loc[(df_inchi_exclusive["topography_change"]) & (~df_inchi_exclusive["constitutional_isomerism"]) & (~df_inchi_exclusive["tautomerism"])  & (~df_inchi_exclusive["change_molecular_formula"]), "change_relative_configuration"] = True

        df_inchi_exclusive["change_absolute_configuration"] = False
        # If chirality barcode is different, but same molecular formula, same heavy atom topology, same topology and same topography then it is an enantiomer
        df_inchi_exclusive.loc[(df_inchi_exclusive["chirality_change"]) & (~df_inchi_exclusive["change_relative_configuration"])  & (~df_inchi_exclusive["constitutional_isomerism"]) & (~df_inchi_exclusive["tautomerism"])  & (~df_inchi_exclusive["change_molecular_formula"]), "change_absolute_configuration"] = True

        for column in columns:
            # Analye if original and duplicate of one column are different
            del df_inchi_exclusive[f"{column}_change"]

            del df_inchi_exclusive[f"{column}_original"]

            del df_inchi_exclusive[f"{column}_duplicate"]

        df_inchi_exclusive.to_pickle(os.path.join(self.results_path, "inchi_exclusive_duplicates.pkl"))

        self.df_inchi_exclusive = df_inchi_exclusive.copy()

        self.n_inchi_exclusive = len(df_inchi_exclusive)

        self._print_molecules_to_dir(df_inchi_exclusive, os.path.join(self.results_path, "inchi_exclusive_duplicates"), "InChI exlusive duplicates")

    
    def _do_statistics_inchi_exclusive_duplicates(self):

        self._analyze_molecular_formula_change()

        self._analyze_tautomerism()

        self._analyze_constitutional_isomerism()

        self._analyze_change_relative_configuration()

        self._analyze_change_absolute_configuration()

        self._get_inchi_exclusive_duplicates_plot()

    def _get_inchi_exclusive_duplicates_plot(self):

        absolute_values = np.array([self.n_molecular_formula_change, self.n_tautomerism, self.n_constitutional_isomerism,  self.n_change_relative_configuration,  self.n_change_absolute_configuration])

        relative_values = np.array([self.p_molecular_formula_change,  self.p_tautomerism,  self.p_constitutional_isomerism,  self.p_change_relative_configuration,  self.p_change_absolute_configuration])

        labels = np.array(["Different \n molecular \n formulas", "Tautomerism", "Consitutional \n isomerism", "Different rel. \n configuration", "Different abs. \n configuration"])

        plt.style.use(self.style_path)

        plt.figure(figsize=(12,9))
        
        plt.bar(0, relative_values[0])
        plt.bar(1, relative_values[1])
        plt.bar(2, relative_values[2])
        plt.bar(3, relative_values[3])
        plt.bar(4, relative_values[4])

        plt.annotate("{:.0f}".format(absolute_values[0]), (0,relative_values[0]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[0]), (0,relative_values[0]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[1]), (1,relative_values[1]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[1]), (1,relative_values[1]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[2]), (2,relative_values[2]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[2]), (2,relative_values[2]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[3]), (3,relative_values[3]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[3]), (3,relative_values[3]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[4]), (4,relative_values[4]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[4]), (4,relative_values[4]+1), ha="center", fontsize=20)

        plt.ylim(0,110)

        plt.yticks(fontsize=20)

        plt.xticks([0,1,2,3, 4], labels, ha="center", fontsize=20)

        plt.ylabel("Share of detected differences by MolBar \n for InChI-exclusive duplicates [%]", fontsize=20)

        plt.tight_layout()
        
        plt.savefig(os.path.join(self.figure_path, "inchi_exclusive_duplicate_analysis.png"), dpi=300, bbox_inches='tight')

    def _analyze_molecular_formula_change(self):

        def get_n_h_atoms(molecula_formula: str):
            """
            Get the number of hydrogen atoms in a molecule.

            Args:
                molecula_formula (str): The molecular formula of the molecule.

            Returns:
                int: The number of hydrogen atoms in the molecule.
            """
            pattern = r'H(\d+)'
            result = re.findall(pattern, molecula_formula)

            try:
                result = int(result[0])

            except IndexError:

                if "H" in molecula_formula:

                    result = 1
                else:

                    result = 0
                
            return result
        
        dff = self.df_inchi_exclusive[self.df_inchi_exclusive["change_molecular_formula"]].copy()

        self.n_molecular_formula_change = dff["change_molecular_formula"].sum()

        self.p_molecular_formula_change = dff["change_molecular_formula"].sum()/self.n_inchi_exclusive*100

        dff["n_H_atoms_original"] = dff["molecular_formula_original"].apply(get_n_h_atoms)

        dff["n_H_atoms_duplicate"] = dff["molecular_formula_duplicate"].apply(get_n_h_atoms)

        dff.loc[:, "n_H_atoms_difference"] = 0

        dff.loc[dff["n_H_atoms_original"] != dff["n_H_atoms_duplicate"], "n_H_atoms_difference"] = 1
        
        self.n_different_h_atoms = dff["n_H_atoms_difference"].sum()


    def _analyze_tautomerism(self):

        dff = self.df_inchi_exclusive[self.df_inchi_exclusive["tautomerism"]]

        self.n_tautomerism = dff["tautomerism"].sum()

        self.p_tautomerism = dff["tautomerism"].sum()/self.n_inchi_exclusive*100

    def _analyze_constitutional_isomerism(self):

        dff = self.df_inchi_exclusive[self.df_inchi_exclusive["constitutional_isomerism"]]

        self.n_constitutional_isomerism = dff["constitutional_isomerism"].sum()

        self.p_constitutional_isomerism = dff["constitutional_isomerism"].sum()/self.n_inchi_exclusive*100

    def _analyze_change_relative_configuration(self):

        dff = self.df_inchi_exclusive[self.df_inchi_exclusive["change_relative_configuration"]]

        self.n_change_relative_configuration = dff["change_relative_configuration"].sum()

        self.p_change_relative_configuration = dff["change_relative_configuration"].sum()/self.n_inchi_exclusive*100

    def _analyze_change_absolute_configuration(self):

        dff = self.df_inchi_exclusive[self.df_inchi_exclusive["change_absolute_configuration"]]

        self.n_change_absolute_configuration = dff["change_absolute_configuration"].sum()

        self.p_change_absolute_configuration = dff["change_absolute_configuration"].sum()/self.n_inchi_exclusive*100



    def _create_report(self):
        
        with open(self.report_path, "w") as file:
            print("#"*97, file=file)
            print("#"*36 + " Duplication Test Report " + "#"*36, file=file)
            print("#"*97, file=file)
            print("Number of molecules in dataset: ", len(self.df_total), file=file)
            print("Number of duplicates due to MolBar: ", self.df_total["duplicate_molbar"].sum(), file=file)
            print("Number of duplicates due to InChI: ", self.df_total["duplicate_inchi"].sum(), file=file)
            print("Number of duplicates due to SMILES: ", self.df_total["duplicate_smiles"].sum(), file=file)
            print("#"*97, file=file)
            print("Number of differences between MolBar and InChI: ", len(self.df_differences), file=file)
            print(f"{len(self.df_molbar_exclusive)} molecules are due to MolBar only duplicates, but InChI finds differences.", file=file)
            print(f"{self.n_inchi_exclusive} molecules are due to InChI only duplicates, but MolBar finds differences because of:", file=file)
            print(f"Different molecular formulae: {self.n_molecular_formula_change} ({self.p_molecular_formula_change:.2f}%)", file=file)
            print(f" \t From those molecular formulae differences: different number of hydrogen atoms: {self.n_different_h_atoms}", file=file)
            print(f"Detected tautomerism: {self.n_tautomerism} ({self.p_tautomerism:.2f}%)", file=file)
            print(f"Detected constitutional isomerism: {self.n_constitutional_isomerism} ({self.p_constitutional_isomerism:.2f}%)", file=file)
            print(f"Detected change in relative configuration of fragments: {self.n_change_relative_configuration} ({self.p_change_relative_configuration:.2f}%)", file=file)
            print(f"Detected change in absolute configuration of fragments: {self.n_change_absolute_configuration} ({self.p_change_absolute_configuration:.2f}%)",  file=file)

    
    def print_report(self):

        with open(self.report_path, "r") as file:

            print(file.read())


    def _animation(self):

        for c in itertools.cycle(['|', '/', '-', '\\']):

            if self.doneloading:
                break
            sys.stdout.write('\rloading ' + c)
            sys.stdout.flush()
            time.sleep(0.1)



    
