import numpy as np
from scipy.stats import rankdata

class FragmentAnalysis:

    def _argsort_fragments(self, atoms_in_fragment_original, atoms_in_fragment_duplicate, orignal_final_energies, duplicate_final_energies):

        charges_in_fragment_original = [sorted(list(self.data_original["atomic_numbers"][atoms])) for atoms in atoms_in_fragment_original]

        charges_in_fragment_original = [ x + [orignal_final_energies[i]] for i, x in enumerate(charges_in_fragment_original)]

        padded_charges_in_fragment_original = self.get_padded_lists(charges_in_fragment_original)

        argsort_original = self.argsort_lists(padded_charges_in_fragment_original)

        charges_in_fragment_duplicate = [sorted(list(self.data_duplicate["atomic_numbers"][atoms])) for atoms in atoms_in_fragment_duplicate]

        charges_in_fragment_duplicate = [ x + [duplicate_final_energies[i]] for i, x in enumerate(charges_in_fragment_duplicate)]

        padded_charges_in_fragment_duplicate = self.get_padded_lists(charges_in_fragment_duplicate)

        argsort_duplicate = self.argsort_lists(padded_charges_in_fragment_duplicate)

        return argsort_original, argsort_duplicate
    
    def _compare_chirality_fragment_wise(self):

        original_fragments_data = list(self.data_original["fragment_data"].values())

        duplicate_fragments_data = list(self.data_duplicate["fragment_data"].values())

        atoms_in_fragments_original = [np.array(data["atoms_in_fragment"])-1 for data in original_fragments_data]

        atoms_in_fragment_duplicate = [np.array(data["atoms_in_fragment"])-1 for data in duplicate_fragments_data]

        original_final_energies = [data["final_energy"] for data in original_fragments_data]

        duplicate_final_energies = [data["final_energy"] for data in duplicate_fragments_data]

        argsort_original, argsort_duplicate = self._argsort_fragments(atoms_in_fragments_original, atoms_in_fragment_duplicate, original_final_energies, duplicate_final_energies)

        causes = []

        for original_index, duplicate_index in zip(argsort_original, argsort_duplicate):

            original_fragment = original_fragments_data[original_index]

            duplicate_fragment = duplicate_fragments_data[duplicate_index]

            atoms_in_fragment_original = np.array(original_fragment["atoms_in_fragment"])-1

            original_elements_in_fragment = np.array(original_fragment["elements"])

            original_chirality_index = original_fragment["absconf_index"]

            duplicate_chirality_index = duplicate_fragment["absconf_index"]

            if original_chirality_index != duplicate_chirality_index:

                cause = self._check_fragment_chirality(original_elements_in_fragment, atoms_in_fragment_original)
                
                if cause is not None:

                    causes.append(cause)

        return causes

    def _check_fragment_chirality(self, original_elements_in_fragment, atoms_in_fragment_original):

        original_nitrogen_atoms = [x for x in np.where(original_elements_in_fragment == "N")[0]]

        global_nitrogen_indices = atoms_in_fragment_original[original_nitrogen_atoms]

        original_poshporous_atoms = [x for x in np.where(original_elements_in_fragment == "P")[0]]

        global_poshporous_indices = atoms_in_fragment_original[original_poshporous_atoms]

        original_sulfur_atoms = [x for x in np.where(original_elements_in_fragment == "S")[0]]

        global_sulfur_indices = atoms_in_fragment_original[original_sulfur_atoms]

        original_carbon_atoms = [x for x in np.where(original_elements_in_fragment == "C")[0]]

        global_carbon_indices = atoms_in_fragment_original[original_carbon_atoms]

        found_cause = False

        cause = None

        if len(original_sulfur_atoms) > 0 and found_cause is False:

            degree = np.array([len(self.data_original["graph"][s]) for s in global_sulfur_indices])

            if np.any(degree >= 3):

                found_cause = True

                cause = "change_of_absolute_configuration_with_sulfur"

        if len(original_poshporous_atoms) > 0 and found_cause is False:

            degree = np.array([len(self.data_original["graph"][p]) for p in global_poshporous_indices])

            if np.any(degree >= 3):

                found_cause = True

                cause = "change_of_absolute_configuration_with_phosphorous"

        if len(original_carbon_atoms) > 0 and found_cause is False:

            degree = np.array([len(self.data_original["graph"][c]) for c in global_carbon_indices])

            if np.any(degree == 3):

                found_cause = True

                cause = "change_of_absolute_configuration_with_carbon"

        return cause

    def _compare_topography_fragment_wise(self):

        original_fragments_data = list(self.data_original["fragment_data"].values())

        duplicate_fragments_data = list(self.data_duplicate["fragment_data"].values())

        atoms_in_fragments_original = [np.array(data["atoms_in_fragment"])-1 for data in original_fragments_data]

        atoms_in_fragment_duplicate = [np.array(data["atoms_in_fragment"])-1 for data in duplicate_fragments_data]

        original_final_energies = [data["final_energy"] for data in original_fragments_data]

        duplicate_final_energies = [data["final_energy"] for data in duplicate_fragments_data]

        argsort_original, argsort_duplicate = self._argsort_fragments(atoms_in_fragments_original, atoms_in_fragment_duplicate, original_final_energies, duplicate_final_energies)

        causes = []

        if len(argsort_original) != len(argsort_duplicate):

            return ["fragmentation_error"]

        for original_index, duplicate_index in zip(argsort_original, argsort_duplicate):

            original_fragment = original_fragments_data[original_index]

            duplicate_fragment = duplicate_fragments_data[duplicate_index]

            atoms_in_fragment_original = np.array(original_fragment["atoms_in_fragment"])-1

            atoms_in_fragment_duplicate = np.array(duplicate_fragment["atoms_in_fragment"])-1

            db_cause, changed_db_atoms = self._check_fragments_for_double_bonds(original_fragment, duplicate_fragment, atoms_in_fragment_original, atoms_in_fragment_duplicate)

            all_changed_db_atoms = set([atom for atoms in changed_db_atoms for atom in atoms])

            if db_cause is not None:

                causes.append(db_cause)

            else:

                relconf_causes = self._check_fragment_topography_spectra(all_changed_db_atoms, original_fragment, duplicate_fragment, atoms_in_fragment_original, atoms_in_fragment_duplicate)

                for cause in relconf_causes:

                    if cause is not None:

                        causes.append(cause)

        return causes

            
    def _check_fragment_topography_spectra(self, changed_db_atoms, original_fragment, duplicate_fragment, atoms_in_fragment_original, atoms_in_fragment_duplicate):

        original_fragment_barcode = self._get_fragment_barcode(self.data_original, original_fragment, atoms_in_fragment_original)

        duplicate_fragment_barcode = self._get_fragment_barcode(self.data_duplicate, duplicate_fragment, atoms_in_fragment_duplicate)

        causes = []

        if original_fragment_barcode != duplicate_fragment_barcode:

            original_elements_in_fragment = np.array(original_fragment["elements"])

            original_nitrogen_atoms = [x for x in np.where(original_elements_in_fragment == "N")[0] if x not in changed_db_atoms]

            global_nitrogen_indices = atoms_in_fragment_original[original_nitrogen_atoms]

            original_poshporous_atoms = [x for x in np.where(original_elements_in_fragment == "P")[0] if x not in changed_db_atoms]

            global_poshporous_indices = atoms_in_fragment_original[original_poshporous_atoms]

            original_sulfur_atoms = [x for x in np.where(original_elements_in_fragment == "S")[0] if x not in changed_db_atoms]

            global_sulfur_indices = atoms_in_fragment_original[original_sulfur_atoms]

            found_cause = False

            if len(original_sulfur_atoms) > 0 and found_cause is False:

                degree = np.array([len(self.data_original["graph"][s]) for s in global_sulfur_indices])

                if np.any(degree >= 3):

                    found_cause = True

                    causes.append("change_of_relative_configuration_with_sulfur")

            if len(original_poshporous_atoms) > 0 and found_cause is False:

                degree = np.array([len(self.data_original["graph"][p]) for p in global_poshporous_indices])

                if np.any(degree >= 3):

                    found_cause = True

                    causes.append("change_of_relative_configuration_with_phosphorous")

            if len(original_nitrogen_atoms) > 0 and found_cause is False:

                degree = np.array([len(self.data_original["graph"][n]) for n in global_nitrogen_indices])

                if np.any(degree == 3):

                    found_cause = True

                    causes.append("change_of_relative_configuration_with_nitrogen")

            return causes

        else:

            return []
        

    def _get_fragment_barcode(self, data_dict, fragment_data, atoms_in_fragment):

        cm_matrix = data_dict["cm_matrix"]

        priorities = data_dict["priorities"]

        local_cm_matrix = cm_matrix[np.ix_(atoms_in_fragment, atoms_in_fragment)]

        local_priorities = rankdata(priorities[atoms_in_fragment], method="min")

        np.fill_diagonal(local_cm_matrix, local_priorities)

        fragment_barcode_int = list(np.round(np.linalg.eigvalsh(local_cm_matrix)*10000).astype(int))

        fragment_barcode_str = " ".join([str(x) for x in fragment_barcode_int])

        return fragment_barcode_str


