import pandas as pd
import numpy as np
import os
import re
import shutil
import matplotlib.pyplot as plt
from tqdm import tqdm
from modules._double_bond_analysis import DoubleBondAnalysis
from modules._fragment_topography_analysis import FragmentAnalysis
from molbar.helper.printer import Printer
from molbar.barcode import get_molbars_from_coordinates
from sklearn.preprocessing import MultiLabelBinarizer

class InChIExclusive(DoubleBondAnalysis, FragmentAnalysis):

    def _analyze_inchi_exclusive_duplicates(self):
        """
        Constructor for the TopographyAnalysis class.
        Reads in the dataframe and adds new columns for the topography analysis.

        Args:
            dataframe_path (_type_): _description_
        """

        inchi_exclusive_dataframe_path = self.results_path + "/inchi_exclusive_duplicates.pkl"

        df_complete = pd.read_pickle(inchi_exclusive_dataframe_path)

        self.df_rel = df_complete[df_complete["change_relative_configuration"]]

        self.df_abs = df_complete[df_complete["change_absolute_configuration"]]

        self.changed_db_types = []

        self.changed_db_wbos = []

        self.changed_db_cid = []

        print("_"*97)
        print("Analyzing InChI-exclusive duplicates for change in relative configuration...")
        self._analyze_topography_inchi_exclusive_duplicates()

        print("_"*97)
        print("Analyzing InChI-exclusive duplicates for change in absolute configuration...")
        self._analyze_chirality_inchi_exclusive_duplicates()

    def _analyze_topography_inchi_exclusive_duplicates(self):

        self._get_molbar_data_rel()

        self._compare_duplicate_entries_topography()

        self._analyze_topography_changes()

        self.get_topography_plot()

        self.get_db_plot()

    def _analyze_chirality_inchi_exclusive_duplicates(self):

        self._get_molbar_data_abs() 

        self._compare_duplicate_entries_chirality()

        self._analyze_chirality_changes()

        self._get_chirality_plot()

    def _get_molbar_data_rel(self):
        """
        Gets the molbar data (with return_data=True) for the original and duplicate entries for later analysis of double bond configuration and topography.
        """

        dff = self.df_rel.copy()

        print("Getting MolBar data for original entries...")

        results = get_molbars_from_coordinates(list_of_coordinates=self.df_rel["coordinates_original"], list_of_elements=self.df_rel["elements_original"], threads=8, progress=True, return_data=True)

        molbars_original = [result[0] for result in results]

        data_original = [result[1] for result in results]

        dff["molbar_data_original"] = data_original.copy()

        dff["molbar_original"] = molbars_original.copy()
        
        print("Getting MolBar data for duplicate entries...")

        results = get_molbars_from_coordinates(list_of_coordinates=self.df_rel["coordinates_duplicate"], list_of_elements=self.df_rel["elements_duplicate"], threads=8, progress=True, return_data=True)

        molbars_duplicate = [result[0] for result in results]

        data_duplicate = [result[1] for result in results]

        dff["molbar_data_duplicate"] = data_duplicate.copy()

        dff["molbar_duplicate"] = molbars_duplicate.copy()

        dff = dff[dff["molbar_original"] != dff["molbar_duplicate"]]

        self.df_rel = dff.copy()

    def _get_molbar_data_abs(self):
        """
        Gets the molbar data (with return_data=True) for the original and duplicate entries for later analysis of double bond configuration and topography.
        """

        dff = self.df_abs.copy()

        print("Getting MolBar data for original entries...")
        results = get_molbars_from_coordinates(list_of_coordinates=self.df_abs["coordinates_original"], list_of_elements=self.df_abs["elements_original"], threads=8, progress=True, return_data=True)

        molbars_original = [result[0] for result in results]

        data_original = [result[1] for result in results]

        dff["molbar_data_original"] = data_original.copy()

        dff["molbar_original"] = molbars_original.copy()

        print("Getting MolBar data for duplicate entries...")
        results = get_molbars_from_coordinates(list_of_coordinates=self.df_abs["coordinates_duplicate"], list_of_elements=self.df_abs["elements_duplicate"], threads=8, progress=True, return_data=True)

        molbars_duplicate = [result[0] for result in results]

        data_duplicate = [result[1] for result in results]

        dff["molbar_data_duplicate"] = data_duplicate.copy()

        dff["molbar_duplicate"] = molbars_duplicate.copy()

        dff = dff[dff["molbar_original"] != dff["molbar_duplicate"]]

        self.df_abs = dff.copy()

    def _prepare_data_for_entry(self, row, type="original"):
        """
        Prepares the data for the original or duplicate entry for later analysis of double bond configuration and topography.

        Args:
            row  (tuple): Row of the dataframe
            type (str, optional): Weither to prepare the data for the original or duplicate entry
        """

        data_dict = {}

        molbar_data = row[1][f"molbar_data_{type}"]

        data_dict["CID"] = row[1][f"CID_{type}"]

        data_dict["coordinates"] = row[1][f"coordinates_{type}"]
    
        data_dict["elements"] = row[1][f"elements_{type}"]

        n_atoms = len(data_dict["elements"])

        data_dict["n_atoms"] = n_atoms

        priorities_dict = molbar_data["priorities"]

        data_dict["priorities"] = np.array([priorities_dict[i+1] for i in range(n_atoms)])

        atomic_numbers_dict = molbar_data["atomic_numbers"]

        data_dict["atomic_numbers"] = np.array([atomic_numbers_dict[i+1] for i in range(n_atoms)])

        graph = self._get_graph(row, type)

        data_dict["graph"] = graph

        data_dict["degrees"] = np.array([len(graph[i]) for i in range(n_atoms)])
        
        data_dict["double_bonds"] = [[bond[0]-1, bond[1]-1] for bond in molbar_data["double_bonds"]]

        data_dict["cm_matrix"] = np.array(molbar_data["unified_coulomb_matrix"])

        data_dict["rings"] = molbar_data["cycles"] 

        data_dict["vsepr_classes"] = list(molbar_data["vsepr_classes"].values())

        data_dict["fragment_data"] = molbar_data["fragment_data"]

        if type == "original":

            self.data_original = data_dict

        else:

            self.data_duplicate = data_dict

    def _compare_duplicate_entries_topography(self):
        """
        Go through the dataframe and compare the original and duplicate entries by comparing the double bond configuration and the topography of the fragments containing the dummy element.

        Args:
            None
        Returns:
            None
        """

        all_causes = []

        cause_columns = ["change_double_bond_configuration", "change_of_relative_configuration_with_nitrogen","change_of_relative_configuration_with_phosphorous","change_of_relative_configuration_with_sulfur","fragmentation_error"]

        print("Comparing duplicate entries...")
        for row in tqdm(self.df_rel.iterrows(), total=len(self.df_rel)):

            self._prepare_data_for_entry(row, type="original")

            self._prepare_data_for_entry(row, type="duplicate")

            causes = self._compare_topography_fragment_wise()

            all_causes.append(causes)

        self.df_rel["causes"] = all_causes

        # Initialize the MultiLabelBinarizer
        mlb = MultiLabelBinarizer()

        # Fit and transform the column containing lists
        one_hot_encoded = pd.DataFrame(mlb.fit_transform(self.df_rel['causes']), columns=mlb.classes_, index=self.df_rel.index)

        # Concatenate the one-hot-encoded columns to the original DataFrame
        self.df_rel = pd.concat([self.df_rel, one_hot_encoded], axis=1)

        # Drop the original column containing lists
        self.df_rel = self.df_rel.drop('causes', axis=1)

        for column in cause_columns:

            if column not in self.df_rel.columns:

                self.df_rel[column] = 0

        self.df_rel["other"] = 0

        mask = (self.df_rel[cause_columns] == 0).all(axis=1)

        self.df_rel.loc[mask, 'other'] = 1

        self.df_rel.to_pickle(os.path.join(self.results_path, "df_rel_configuration.pkl"))

    def _compare_duplicate_entries_chirality(self):
        """
        Go through the dataframe and compare the original and duplicate entries by comparing the chirality of the fragments containing the dummy element.

        Args:
            None
        Returns:
            None
        """

        all_causes = []

        cause_columns = ["change_of_absolute_configuration_with_phosphorous","change_of_absolute_configuration_with_sulfur", "change_of_absolute_configuration_with_carbon"]

        print("Comparing duplicate entries...")
        for row in tqdm(self.df_abs.iterrows(), total=len(self.df_abs)):

            self._prepare_data_for_entry(row, type="original")

            self._prepare_data_for_entry(row, type="duplicate")

            causes = self._compare_chirality_fragment_wise()

            all_causes.append(causes)

        self.df_abs["causes"] = all_causes

        # Initialize the MultiLabelBinarizer
        mlb = MultiLabelBinarizer()

        # Fit and transform the column containing lists
        one_hot_encoded = pd.DataFrame(mlb.fit_transform(self.df_abs['causes']), columns=mlb.classes_, index=self.df_abs.index)

        # Concatenate the one-hot-encoded columns to the original DataFrame
        self.df_abs = pd.concat([self.df_abs, one_hot_encoded], axis=1)

        # Drop the original column containing lists
        self.df_abs = self.df_abs.drop('causes', axis=1)

        for column in cause_columns:

            if column not in self.df_abs.columns:

                self.df_abs[column] = 0

        self.df_abs["other"] = 0

        mask = (self.df_abs[cause_columns] == 0).all(axis=1)

        self.df_abs.loc[mask, 'other'] = 1

        self.df_abs.to_pickle(os.path.join(self.results_path, "df_abs_configuration.pkl"))


    def _analyze_chirality_changes(self):

        n_total = len(self.df_abs)

        self.n_change_of_absolute_configuration_with_phosphorous = self.df_abs["change_of_absolute_configuration_with_phosphorous"].sum()

        self.p_change_of_absolute_configuration_with_phosphorous = self.n_change_of_absolute_configuration_with_phosphorous/n_total*100

        self.n_change_of_absolute_configuration_with_sulfur = self.df_abs["change_of_absolute_configuration_with_sulfur"].sum()

        self.p_change_of_absolute_configuration_with_sulfur = self.n_change_of_absolute_configuration_with_sulfur/n_total*100

        self.n_change_of_absolute_configuration_with_carbon = self.df_abs["change_of_absolute_configuration_with_carbon"].sum()

        self.p_change_of_absolute_configuration_with_carbon = self.n_change_of_absolute_configuration_with_carbon/n_total*100

        self.n_unknown = self.df_abs["other"].sum()

        self.p_unknown = self.n_unknown/n_total*100

        dff = self.df_abs[self.df_abs["other"] == 0].copy()

        self.df_abs = dff.copy()

        with open(self.report_path, "a") as report:
            print("#"*97, file=report)
            print("Report on the analysis of absolute chirality changes in InChI-exclusive duplicates", file=report)
            print("Number of cases: ", len(self.df_abs), file=report)
            print("#"*97, file=report)
            print("Change of absolute configuration of fragments:", file=report)
            print("\t Number of fragments with Sulphur", self.n_change_of_absolute_configuration_with_sulfur, "({:.2f}%)".format(self.p_change_of_absolute_configuration_with_sulfur), file=report)
            print("\t Number of fragments with Phosphorus", self.n_change_of_absolute_configuration_with_phosphorous, "({:.2f}%)".format(self.p_change_of_absolute_configuration_with_phosphorous), file=report)
            print("\t Number of fragments with Carbon", self.n_change_of_absolute_configuration_with_carbon, "({:.2f}%)".format(self.p_change_of_absolute_configuration_with_carbon), file=report)

    def _analyze_topography_changes(self):

        n_total = len(self.df_rel)

        self.n_change_double_bond_configuration	= self.df_rel["change_double_bond_configuration"].sum()

        self.p_change_double_bond_configuration = self.n_change_double_bond_configuration/n_total*100
        
        self.n_change_of_relative_configuration_with_nitrogen = self.df_rel["change_of_relative_configuration_with_nitrogen"].sum()

        self.p_change_of_relative_configuration_with_nitrogen = self.n_change_of_relative_configuration_with_nitrogen/n_total*100

        self.n_change_of_relative_configuration_with_phosphorous = self.df_rel["change_of_relative_configuration_with_phosphorous"].sum()

        self.p_change_of_relative_configuration_with_phosphorous = self.n_change_of_relative_configuration_with_phosphorous/n_total*100

        self.n_change_of_relative_configuration_with_sulfur = self.df_rel["change_of_relative_configuration_with_sulfur"].sum()

        self.p_change_of_relative_configuration_with_sulfur = self.n_change_of_relative_configuration_with_sulfur/n_total*100

        self.n_fragmentation_error = self.df_rel["fragmentation_error"].sum()

        self.p_fragmentation_error = self.n_fragmentation_error/n_total*100

        self.n_unknown = self.df_rel["other"].sum()

        self.p_unknown = self.n_unknown/n_total*100

        with open(self.report_path, "a") as report:
            print("#"*97, file=report)
            print("Report on the analysis of topography changes in InChI-exclusive duplicates", file=report)
            print("#"*97, file=report)
            print("Number of cases: ", len(self.df_rel), file=report)
            print("Change of topography due to different configuration of double bonds:", self.n_change_double_bond_configuration, "({:.2f}%)".format(self.p_change_double_bond_configuration), file=report)
            print("_"*97, file=report)
            print("Change of topography due to different relative configuration:")
            print("\t Number of fragments with Sulphur", self.n_change_of_relative_configuration_with_sulfur, "({:.2f}%)".format(self.p_change_of_relative_configuration_with_sulfur), file=report)
            print("\t Number of fragments with Phosphorus", self.n_change_of_relative_configuration_with_phosphorous, "({:.2f}%)".format(self.p_change_of_relative_configuration_with_phosphorous), file=report)
            print("\t Number of fragments with Nitrogen", self.n_change_of_relative_configuration_with_nitrogen, "({:.2f}%)".format(self.p_change_of_relative_configuration_with_nitrogen), file=report)
            print("_"*97, file=report)
            print("Changes in fragmentation error (different number of fragments)", self.n_fragmentation_error, "({:.2f}%)".format(self.p_fragmentation_error), file=report)
            print("_"*97, file=report) 
            print("Changes in topography that could not be classified", self.n_unknown, "({:.2f}%)".format(self.p_unknown), file=report)


    def _get_graph(self, row, type="original"):
        """
        Construct the graph of the original or duplicate entry.

        Args:
            row (tuple): Row of the dataframe
            type (str, optional): Wether to construct the graph of the original or duplicate entry

        Returns:
            graph (dict): Graph of the original or duplicate entry
        """

        data = row[1][f"molbar_data_{type}"]

        n_atoms = len(data["elements"])

        single_bonds = data["single_bonds"]

        double_bonds = data["double_bonds"]

        triple_bonds = data["triple_bonds"]

        graph = {key: [] for key in range(n_atoms)}

        for bond in single_bonds:

            graph[bond[0]-1].append(bond[1]-1)
            graph[bond[1]-1].append(bond[0]-1)
        
        for bond in double_bonds:

            graph[bond[0]-1].append(bond[1]-1)
            graph[bond[1]-1].append(bond[0]-1)

        for bond in triple_bonds:

            graph[bond[0]-1].append(bond[1]-1)
            graph[bond[1]-1].append(bond[0]-1)

        return graph
    
    def get_padded_lists(self, list_of_lists):

        max_length = max([len(single_list) for single_list in list_of_lists])

        padded_list = [single_list + [0]*(max_length-len(single_list)) for single_list in list_of_lists]

        return padded_list
    
    
    def argsort_lists(self, padded_list_of_lists):

        sorted_priorities = sorted(padded_list_of_lists, key=lambda x: tuple([x[i] for i in range(len(x))]), reverse=True)

        sorted_indices = []

        for sorted_priority in sorted_priorities:

            indices = np.argwhere([True if priority == sorted_priority else False for priority in padded_list_of_lists])

            for index in indices:

                if index not in sorted_indices:

                    sorted_indices.append(index[0])

        return sorted_indices
    
    def _get_chirality_plot(self):
        """
        Generate a plot of the number of duplicates in the dataset.

        Parameters:
        -----------
        df: pandas.DataFrame

        Returns:
        --------
        None
        """

        plt.style.use(self.style_path)

        absolute_values = np.array([self.n_change_of_absolute_configuration_with_sulfur, self.n_change_of_absolute_configuration_with_phosphorous, self.n_change_of_absolute_configuration_with_carbon]) 
                                    
        relative_values = np.array([self.p_change_of_absolute_configuration_with_sulfur, self.p_change_of_absolute_configuration_with_phosphorous, self.p_change_of_absolute_configuration_with_carbon])

        labels = np.array(["Sulfur", "Phosphorous", "Carbon"])
                           
        sorted_indices = np.argsort(absolute_values)

        sorted_indices = sorted_indices[::-1]

        absolute_values = absolute_values[sorted_indices]

        relative_values = relative_values[sorted_indices]

        labels = labels[sorted_indices]

        plt.figure(figsize=(12,9))
        plt.bar(0, relative_values[0])
        plt.bar(1, relative_values[1])
        plt.bar(2, relative_values[2])

        plt.annotate("{:.0f}".format(absolute_values[0]), (0,relative_values[0]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[0]), (0,relative_values[0]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[1]), (1,relative_values[1]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[1]), (1,relative_values[1]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[2]), (2,relative_values[2]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[2]), (2,relative_values[2]+1), ha="center", fontsize=20)

        plt.ylim(0,110)

        plt.xticks([0,1,2], labels, ha="center", fontsize=20)

        plt.ylabel("Share of classified absolute configuration differences \n by MolBar for InChI-exclusive duplicates [%]", fontsize=20)

        plt.tight_layout()
        
        plt.savefig(os.path.join(self.figure_path, "inchi_exclusive_duplicate_absolute_configuration_analysis.png"), dpi=300, bbox_inches='tight')
    
    def get_topography_plot(self):
        """
        Generate a plot of the number of duplicates in the dataset.

        Parameters:
        -----------
        df: pandas.DataFrame

        Returns:
        --------
        None
        """
        
        plt.style.use(self.style_path)

        absolute_values = np.array([self.n_change_double_bond_configuration, self.n_change_of_relative_configuration_with_nitrogen, self.n_change_of_relative_configuration_with_phosphorous, self.n_change_of_relative_configuration_with_sulfur, self.n_unknown])

        relative_values = np.array([self.p_change_double_bond_configuration, self.p_change_of_relative_configuration_with_nitrogen, self.p_change_of_relative_configuration_with_phosphorous, self.p_change_of_relative_configuration_with_sulfur, self.p_unknown])

        labels = np.array(["E–Z config. \n double bonds", "Nitrogen", "Phosphorous", "Sulfur", "Other"])
                           
        sorted_indices = np.argsort(absolute_values)

        sorted_indices = sorted_indices[::-1]

        absolute_values = absolute_values[sorted_indices]

        relative_values = relative_values[sorted_indices]

        labels = labels[sorted_indices]

        plt.figure(figsize=(12,9))
        plt.bar(0, relative_values[0])
        plt.bar(1, relative_values[1])
        plt.bar(2, relative_values[2])
        plt.bar(3, relative_values[3])
        plt.bar(4, relative_values[4])


        plt.annotate("{:.0f}".format(absolute_values[0]), (0,relative_values[0]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[0]), (0,relative_values[0]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[1]), (1,relative_values[1]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[1]), (1,relative_values[1]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[2]), (2,relative_values[2]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[2]), (2,relative_values[2]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[3]), (3,relative_values[3]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[3]), (3,relative_values[3]+1), ha="center", fontsize=20)
        plt.annotate("{:.0f}".format(absolute_values[4]), (4,relative_values[4]+6), ha="center", fontsize=20)
        plt.annotate("({:.1f}%)".format(relative_values[4]), (4,relative_values[4]+1), ha="center", fontsize=20)
 
        plt.ylim(0,110)

        plt.xticks([0,1,2,3,4], labels, ha="center", fontsize=20)

        plt.ylabel("Share of classified topography differences by MolBar \n for InChI-exclusive duplicates [%]", fontsize=20)

        plt.tight_layout()
        
        plt.savefig(os.path.join(self.figure_path, "inchi_exclusive_duplicate_relative_configuration_analysis.png"), dpi=300, bbox_inches='tight')


    def _print_molecules_to_dir(self, df, path, title):

        if os.path.exists(path):

            shutil.rmtree(path)

        os.mkdir(path)

        print(f"Printing {title} to", path)

        for row in tqdm(df.iterrows(), total=len(df)):

            self._print_molfile(path, row)
            #self._print_xyz(path, row)

    def _print_molfile(self, path, row):

        file_original = row[1]["molfile_original"]

        CID_original = row[1]["CID_original"]

        file_duplicate = row[1]["molfile_duplicate"]

        CID_duplicate = row[1]["CID_duplicate"]

        with open(f"{path}/{CID_original}.mol", "w") as f:

            f.write(file_original)

        with open(f"{path}/{CID_duplicate}.mol", "w") as f:

            f.write(file_duplicate)

    def _print_xyz(self, path, row):

        #file_original = self.get_file_for_molecule(row[1]["comment_original"])

        CID_original = row[1]["CID_original"]

        #file_duplicate = self.get_file_for_molecule(row[1]["comment_duplicate"])

        CID_duplicate = row[1]["CID_duplicate"]

        n_atoms_original = len(row[1]["elements_original"])

        elements_original = row[1]["elements_original"]

        coordinates_original = row[1]["coordinates_original"]

        elements_duplicate = row[1]["elements_duplicate"]

        coordinates_duplicate = row[1]["coordinates_duplicate"]

        n_atoms_duplicate = len(row[1]["elements_duplicate"])

        Printer(n_atoms_original, "", coordinates_original, elements_original, f"{path}/{CID_original}.xyz").print()

        Printer(n_atoms_duplicate, "", coordinates_duplicate, elements_duplicate, f"{path}/{CID_duplicate}.xyz").print()

        #with open(f"{path}/{CID_original}.sdf", "w") as f:

        #    for line in file_original:

        #        f.write(line)

        #with open(f"{path}/{CID_duplicate}.sdf", "w") as f:

        #    for line in file_duplicate:

        #        f.write(line)

    def _get_file_for_molecule(self, string):

        """
        Prepares the data for the dataset.

        Args:

            datafile: The path to the datafile.

        Returns:

            data: A list of lists containing the data for each molecule.
        """

        count = 0

        for datafile in ["./data/combined_mols_0_to_1000000.sdf", "./data/combined_mols_1000000_to_2000000.sdf", "./data/combined_mols_2000000_to_3000000.sdf", "./data/combined_mols_3000000_to_3899647.sdf"]:

            with open(datafile) as f:

                file = []
                this_is_it = False

                for line in f:

                    if string[:30] in line:

                        this_is_it = True
                    
                    if line == '$$$$\n':

                        if this_is_it:

                            return file

                        count += 1

                        file = []

                    else:

                        file.append(line)

if __name__ == "__main__":

    dataframe_path = "/Users/nilsvanstaalduinen/papermolbar/data/data/duplicate_inchi.pkl"

    duplicate_test = InChIExclusive(dataframe_path)



    

    
